/**********************************************************************
*  ______   ________   ___   __    ________   ___   ___   ______      *
* /_____/\ /_______/\ /__/\ /__/\ /_______/\ /___/\/__/\ /_____/\     *      
* \:::_ \ \\::: _  \ \\::\_\\  \ \\::: _  \ \\::.\ \\ \ \\:::_ \ \    *   
*  \:(_) \ \\::(_)  \ \\:. `-\  \ \\::(_)  \ \\:: \/_) \ \\:\ \ \ \   * 
*   \: ___\/ \:: __  \ \\:. _    \ \\:: __  \ \\:. __  ( ( \:\ \ \ \  * 
*    \ \ \    \:.\ \  \ \\. \`-\  \ \\:.\ \  \ \\: \ )  \ \ \:\_\ \ \ * 
*     \_\/     \__\/\__\/ \__\/ \__\/ \__\/\__\/ \__\/\__\/  \_____\/ *
*                                                                     *
***********************************************************************
*                                                                     *
*                              Panako                                 * 
*                       Acoustic Fingerprinting                       *
*                                                                     *
***********************************************************************/

package be.panako.strategy.fft;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;

import be.hogent.tarsos.dsp.AudioEvent;
import be.hogent.tarsos.dsp.AudioProcessor;
import be.panako.util.Config;
import be.panako.util.Key;


/**
 * Processes audio to store landmarks in an in memory hash, or to persistent storage.
 * 
 * @author Joren Six
 */
public class FFTFingerprintProcessor implements AudioProcessor {	
	/**
	 * Target number of hashes per second of audio
	 */
	private final int numberOfFingerprintsPerSecond;
	
	private final int divideSteps;
	
	private int timeOffset; 
	
	
	/**
	 * A set of unique fingerprints. Ordered by time, last fingerprint first.
	 */
	private final TreeSet<FFTFingerprint> fingerprintTree;

	private int currentStep;
	
	/**

	 */
	public FFTFingerprintProcessor(int numberOfFingerprintsPerSecond,int divideSteps){
		this.numberOfFingerprintsPerSecond = numberOfFingerprintsPerSecond;
		this.divideSteps = 1;
		//order by time
		fingerprintTree = new TreeSet<FFTFingerprint>(new Comparator<FFTFingerprint>(){
			@Override
			public int compare(FFTFingerprint o1, FFTFingerprint o2) {
				return Integer.valueOf(o2.t1).compareTo(o1.t1);
			}});
	}
	
	@Override
	public boolean process(AudioEvent audioEvent) {
		float sampleRate = audioEvent.getSampleRate();
		int stepSize = audioEvent.getBufferSize();
		int fftHopSizeInSamples = Config.getInt(Key.FFT_STEP_SIZE);
		timeOffset = currentStep * stepSize / fftHopSizeInSamples;
		currentStep++;
		float[] buffer = audioEvent.getFloatBuffer();
		Set<FFTFingerprint> fingerprints = FFTFingerprintExtractorNaive.calculateLandmarks(buffer, sampleRate, numberOfFingerprintsPerSecond,timeOffset);
		fingerprintTree.addAll(fingerprints);
		
		if(divideSteps > 1){
			//divide audio block of 32ms in x steps (for better recall)
			int step = fftHopSizeInSamples / divideSteps;
			for(int i = 1; i < divideSteps ; i++){
				buffer = Arrays.copyOfRange(buffer,step,buffer.length);
				fingerprints = FFTFingerprintExtractor.calculateLandmarks(buffer, sampleRate, numberOfFingerprintsPerSecond,timeOffset);
				fingerprintTree.addAll(fingerprints);
			}
		}
		return true;
	}
	
	@Override
	public void processingFinished() {
	}
	
	public Set<FFTFingerprint> getFingerprints(){
		return fingerprintTree;
	}
	
	public int getTimeOffset(){
		return timeOffset;
	}
}
