package be.panako.strategy.fft;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.UnsupportedAudioFileException;

import be.hogent.tarsos.dsp.AudioDispatcher;
import be.hogent.tarsos.dsp.AudioEvent;
import be.hogent.tarsos.dsp.AudioFile;
import be.hogent.tarsos.dsp.AudioProcessor;
import be.panako.strategy.QueryResult;
import be.panako.strategy.QueryResultHandler;
import be.panako.strategy.Strategy;
import be.panako.strategy.fft.storage.FFTFingerprintQueryMatch;
import be.panako.strategy.fft.storage.FFTMapDBStorage;
import be.panako.util.Config;
import be.panako.util.FileUtils;
import be.panako.util.Key;

public class FFTStrategy extends Strategy{
	private final static Logger LOG = Logger.getLogger(FFTStrategy.class.getName());
	
	private final FFTMapDBStorage storage;
	
	public FFTStrategy(){
		storage = FFTMapDBStorage.getInstance();
	}
	
	public double store(String resource, final String description) {
		final int identifier = FileUtils.getIdentifier(resource);
		final double durationInSeconds;
		LOG.info("Store " + description + "...");
		if(storage.hasDescription(description)){
			LOG.warning("Skipping " + description + ", it is already present in the database");
			durationInSeconds = -1L;
		}else{
			final FFTFingerprintProcessor rhap = new FFTFingerprintProcessor(Config.getInt(Key.FFT_FINGERPRINTS_PER_SECOND_FOR_STORAGE),1);
			durationInSeconds = AudioDispatcher.determineDuration(new File(resource));
			AudioDispatcher d = getAudioDispatcher(resource,durationInSeconds);
			d.addAudioProcessor(rhap);
			d.addAudioProcessor(new AudioProcessor() {
					@Override
					public boolean process(AudioEvent audioEvent) {
						for(FFTFingerprint fingerprint: rhap.getFingerprints()){
							storage.addFingerprint(identifier, fingerprint.t1+rhap.getTimeOffset(), fingerprint.hash());
						}
						return true;
					}
					@Override
					public void processingFinished() {
						// Store the meta data.
						storage.addAudio(identifier, description);
						// Commit the changes to store the fingerprints
						storage.audioObjectAdded((int) Math.round(durationInSeconds));
					}
				});
			d.run();
		}
		return durationInSeconds;
	}
	
	public void query(String query, final int maxNumberOfResults, QueryResultHandler handler){
	
		final FFTMapDBStorage storage = FFTMapDBStorage.getInstance();
		
		final List<FFTFingerprintQueryMatch> queryMatches = new ArrayList<FFTFingerprintQueryMatch>();
		
		final double durationInSeconds;
		LOG.info("Query " + query + "...");
		
		final FFTFingerprintProcessor rhap = new FFTFingerprintProcessor(Config.getInt(Key.FFT_LANDMARKS_PER_SECOND_FOR_QUERY),4);
		durationInSeconds = AudioDispatcher.determineDuration(new File(query));
		AudioDispatcher d = getAudioDispatcher(query,durationInSeconds);
		d.addAudioProcessor(rhap);
		d.addAudioProcessor(new AudioProcessor() {
				@Override
				public boolean process(AudioEvent audioEvent) {
					System.out.println(audioEvent.getTimeStamp());
					return true;
				}
				@Override
				public void processingFinished() {
					List<FFTFingerprint> fingerprints = new ArrayList<FFTFingerprint>(rhap.getFingerprints());
					queryMatches.addAll(storage.getMatches(fingerprints, maxNumberOfResults));
				}
			});
		d.run();
		
		if(queryMatches.isEmpty()){
			System.out.println(String.format("Query %s matches NOTHING",query));
		}else{
			for(FFTFingerprintQueryMatch match : queryMatches){
				String description = storage.getAudioDescription(match.identifier);
				handler.handleQueryResult(new QueryResult(String.valueOf(match.identifier), description, match.score, match.getStartTime(),100.0,100.0));
			}
		}
	}

	@Override
	public boolean storageIsAvailable() {
		return FFTMapDBStorage.getInstance() != null;
	}
	
	@Override
	public boolean hasResource(String resource) {
		return FFTMapDBStorage.getInstance().hasDescription(new File(resource).getName());
	}
	
	private AudioDispatcher getAudioDispatcher(String resource, double durationInSeconds){
		AudioDispatcher d = null;
		
		int sampleRate = Config.getInt(Key.FFT_SAMPLE_RATE);
		AudioInputStream stream = new AudioFile(resource).getMonoStream(sampleRate);
		
		// Five minutes fits in memory comfortably and contains a lot of songs.
		int fiveMinutesInSamples = (5*60) * sampleRate;
		
		// The buffer size should be a multiple of the FFT step size
		int totalSamples = (int) Math.round(durationInSeconds * sampleRate);
		int stepSizeSamples = Config.getInt(Key.FFT_STEP_SIZE);
		assert fiveMinutesInSamples % stepSizeSamples == 0;
		
		int stepSize = Math.min(fiveMinutesInSamples, totalSamples);
		
		//set the step size to five minutes  and overlap to zero
		//if the song is shorter than five minutes, use that as the step size
		try {
			d = new AudioDispatcher(stream, stepSize, 0);
		} catch (UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
		return d;
	}

	@Override
	public void printStats() {
		FFTMapDBStorage storage = FFTMapDBStorage.getInstance();
		System.out.println("Number of audio objects: " + storage.getNumberOfAudioObjects());
		System.out.println("Number of fingerprints: " + storage.getNumberOfFingerprints());
		System.out.println("Number of seconds: " + storage.getNumberOfSeconds());
		System.out.println("Number of fingerprints/second: " + storage.getNumberOfFingerprints()/storage.getNumberOfSeconds());
	}	
}
